from .common import *
from ..models.rtsm import RTSMObservation, RTSMError, RTSMSpectrum
from ..serializers.rtsm import RTSMObservationSerializer, RTSMErrorSerializer, \
    RTSMSpectrumSerializer, RTSMSummaryPlotSerializer, RTSMErrorSummary
from ..rtsm_test_raw_parser import parse_rtsm_test
from django.shortcuts import get_object_or_404
from django.http import Http404, HttpResponseServerError
import os
from ..tasks import check_error_summary_plot

logger = logging.getLogger('views')


class RTSMSpectrumViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = RTSMSpectrum.objects.all()
    serializer_class = RTSMSpectrumSerializer
    filter_fields = ['rtsm_error']


class RTSMErrorsViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = RTSMError.objects.all()
    serializer_class = RTSMErrorSerializer
    filter_fields = '__all__'


class RTSMObservationViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = RTSMObservation.objects.all()
    serializer_class = RTSMObservationSerializer
    filter_fields = '__all__'


class RTSMSummaryPlot(viewsets.ViewSet):
    """
    Get the summary plot associated to the error summary given
    """
    queryset = RTSMErrorSummary.objects.all()

    def retrieve(self, request, pk=None):
        try:
            entity = get_object_or_404(self.queryset, pk=pk)
            summary_plot = entity.summary_plot.first()

            if summary_plot is None:
                raise ObjectDoesNotExist()

            uri = RTSMSummaryPlotSerializer(summary_plot).data['uri']

        except ObjectDoesNotExist as e:
            check_error_summary_plot.delay(pk)
            raise Http404()

        if uri and os.path.exists(uri) and os.path.isfile(uri):
                with open(uri, 'rb') as f_stream:
                    image = f_stream.read()
                return HttpResponse(image, status=status.HTTP_200_OK, content_type='image/gif')
        else:
            check_error_summary_plot.delay(pk)
            print(uri)
            raise Http404()


@api_view(['POST'])
def insert_raw_rtsm_test(request):
    """
    This function is meant to parse a request of the form
    {
    "content": "[RTSM TEST RAW TEXT]" station_name: "station_name"
    }
    parse the content field and create all the rtsm_test entity related into the database
    :param request: HTTP request
    :return:
    """
    if request.method == 'POST':

        if 'content' in request.data:
            try:
                logger.debug('handling raw request for %s', request)
                content = request.data['content']
                station_name = request.data['station_name']
                logger.debug('handling raw request for data %s', content)
                entry = parse_rtsm_test(content)
                entry.update(station_name=station_name)

                logger.info('RTSM parsed successfully for obsid %d and station %s',
                            entry['observation_id'],
                            entry['station_name'])

                try:
                    rtsm_item = RTSMObservationSerializer().create(dict(entry))
                    rtsm_id = rtsm_item.id
                except ItemAlreadyExists as e:
                    return Response(status=status.HTTP_200_OK,
                                    data='RTSM test was already present with id {}\n'.format(e.instance_id))
                logger.info('RTSM inserted successfully for obsid %d and station %s',
                            entry['observation_id'],
                            entry['station_name'])
            except KeyError as e:
                logger.exception("raw station info malformed %s: %s",
                                 request.data['content'], e)
                logger.info('station test malformed skipping insertion.')
                return Response(status=status.HTTP_406_NOT_ACCEPTABLE, data='skipping insertion RTSM malformed')
            except Exception as e:
                logger.exception("exception occurred while parsing raw station info %s: %s",
                                 request.data['content'], e)
                return Response(exception=True,
                                data="the post message is not correct." +
                                     " It has to be of the form {content:[RAWSTRING], station_name:[STATION_NAME]}: %s. Request provided %s" % (
                                         e, request),
                                status=status.HTTP_400_BAD_REQUEST)
        else:
            return Response(exception=True,
                            data="the post message is not correct." +
                                 " It has to be of the form \{'content':[RAWSTRING]\}",
                            status=status.HTTP_400_BAD_REQUEST)
        logger.info('request processed correctly inserted')

        return Response(status=status.HTTP_200_OK,
                        data='RTSM test has been inserted with id {}\n'.format(rtsm_id))
